
#include "ACMachine.hpp"


ACPMMachineState::ACPMMachineState() : character_( L'\0' ), parent_( NULL ), failure_( NULL )
{
} 

ACPMMachineState::ACPMMachineState( const ACPMMachineState* parent, wchar_t character ) :
    character_( character ), parent_( parent ), failure_( NULL )
{
} 


ACPMMachine::ACPMMachine() :
    root_( new ACPMMachineState() ), currentState_( NULL )
{
}

void ACPMMachine::addPattern ( const std::wstring& pattern, Expression* associatedObject )
{
    ACPMMachineState* currentState = root_;
    int character = 0;

    const int patternSize = pattern.length();

    ACPMMachineState* newState = FindTransition( currentState, pattern[ character ] );
    while ( newState != NULL && character < patternSize ) {
        currentState = newState; ++character;
        newState = FindTransition( currentState, pattern[ character ] );
    }

    while ( character < patternSize ) {
        newState = new ACPMMachineState ( currentState, pattern[ character ] );
        AddTransition( currentState, newState );
        currentState = newState; ++character;
    }

    currentState->output_.insert( std::make_pair( pattern, associatedObject ) );
}

void ACPMMachine::computeStates() 
{
    StatesQueue statesQueue;

    // set failure transitions to initial state for all states with depth 1
    for (
        StatesHash::const_iterator transition = root_->transitions_.begin();
        transition != root_->transitions_.end(); ++transition
        ) {
            statesQueue.push( transition->second );
            transition->second->failure_ = root_;
    }

    // compute failure transitions for all other depths 
    while ( !statesQueue.empty() ) 
    {
        ACPMMachineState* currentState = statesQueue.front(); statesQueue.pop();

        ACPMMachineState* failureState = currentState->parent_->failure_;
        while ( failureState != NULL && FindTransition( failureState, currentState->character_ ) == NULL )
            failureState = failureState->failure_;

        if ( failureState != NULL ) {
            currentState->failure_ = FindTransition( failureState, currentState->character_ );
            currentState->output_.insert (
                currentState->failure_->output_.begin() , currentState->failure_->output_.end()
                );
        } else {
            currentState->failure_ = root_;
        }

        // put all transitions of current state for failure computing queue
        for (
            StatesHash::const_iterator transition = currentState->transitions_.begin();
            transition != currentState->transitions_.end(); ++transition
            )
        {
                statesQueue.push( transition->second );
        }
    }

    root_->failure_ = root_;
    currentState_ = root_;
} // ACPMMachine::computeStates

const MatchList& ACPMMachine::getCurrentMatch() const
{
    return currentState_->output_;
} // ACPMMachine::getCurrentMatch

void ACPMMachine::reset () 
{
    currentState_ = root_;
} // ACPMMachine::reset

bool ACPMMachine::feedCharacter ( wchar_t character) 
{
    ACPMMachineState* newState = FindTransition( currentState_, character );

    while ( newState == NULL ) {
        if ( currentState_ == root_ ) break;
        currentState_ = currentState_->failure_;
        newState = FindTransition( currentState_, character );
    }

    if ( newState == NULL )
        return false;

    currentState_ = newState;

    return !currentState_->output_.empty();
}

ACPMMachineState* ACPMMachine::FindTransition ( ACPMMachineState* state, wchar_t character ) 
{
    StatesHash::const_iterator it = state->transitions_.find( character );
    if ( it == state->transitions_.end() ) return NULL;
    return it->second;
}

void ACPMMachine::AddTransition ( ACPMMachineState* from, ACPMMachineState* to ) 
{
    from->transitions_[ to->character_ ] = to;
    to->parent_ = from;
}

ACPMMachine::~ACPMMachine() 
{
    StatesQueue statesQueue;
    statesQueue.push( root_ );

    while ( !statesQueue.empty() ) 
    {
        ACPMMachineState* currentState = statesQueue.front();
        statesQueue.pop();

        for (StatesHash::const_iterator transition = currentState->transitions_.begin();
            transition != currentState->transitions_.end();
            ++transition) 
        {
                statesQueue.push( transition->second );
        }

        delete currentState;
    }
} // ACPMMachine::~ACPMMachine
