//
// expression.cpp
//
#include <cctype>
#include <algorithm>
#include <iterator>

#include "expressions.hpp"

// class NotExpression

const std::wstring NotExpression::name ( L"NOT" );

NotExpression::NotExpression ( Expression* expression ) : child_( expression ) {
   assert( child_ != NULL );
} // NotExpression::NotExpression

const std::wstring& NotExpression::sought() const {
   return NotExpression::name;
} // NotExpression::sought

bool NotExpression::evaluate(const stream_list& pos_list_) const {
   assert( child_ != NULL );
   return !child_->evaluate(pos_list_);
} // NotExpression::evaluate

MatchType NotExpression::match(const stream_list& pos_list_) const
{
    return MatchType();
}

void NotExpression::match(const stream_list& pos_list_, MatchType & result_pos_list) const
{
    MatchType child_pos_list;
    child_->match(pos_list_, child_pos_list);
    if (pos_list_.size() == 1 && child_pos_list.size() > 0)
    {
        unsigned long long start_pos = pos_list_.at(0)._start_pos;
        unsigned long long end_pos = pos_list_.at(0)._end_pos;

        unsigned long long pos = start_pos;
        for (MatchType::const_iterator iter = child_pos_list.begin(); 
            iter != child_pos_list.end(); ++iter)
        {
            if (iter->_pos > end_pos) break;
            if (iter->_pos + iter->_size < pos) continue;
            if (iter->_pos <= pos)
            {
                pos = iter->_pos + iter->_size;
                continue;
            }

            unsigned long size = iter->_pos - pos;
            if (pos + size > end_pos)
            {
                size = end_pos - pos;
            }
            Matched not_match(pos, size);
            result_pos_list.push_back(not_match);
            pos = iter->_pos + iter->_size;            
        }
        Matched last_pos = *(child_pos_list.rbegin());
        pos = last_pos._pos + last_pos._size;
        if (pos <= end_pos)
        {
            Matched not_match(pos, end_pos - pos + 1);
            result_pos_list.push_back(not_match);
        }
    }
}

std::size_t NotExpression::childrenCount() const {
   return 1;
} // NotExpression::childrenCount

Expression* NotExpression::child ( std::size_t index ) {
   return index == 0 ? child_ : NULL;
} // NotExpression::child

Expression::Kind::Type NotExpression::kind() const {
   return Expression::Kind::Not;
} // NotExpression::kind

void NotExpression::reset() {
   assert( child_ != NULL );
   child_->reset();
} // NotExpression::reset

NotExpression::~NotExpression() {
   assert( child_ != NULL );
   delete child_;
} // NotExpression::~NotExpression

// class AndExpression

const std::wstring AndExpression::name ( L"AND" );

AndExpression::AndExpression ( Expression* left, Expression* right )
   : left_( left ), right_( right )
{
   assert( left_!= NULL && right_ != NULL );
} // AndExpression::AndExpression

const std::wstring& AndExpression::sought() const {
   return AndExpression::name;
} // AndExpression::sought

bool AndExpression::evaluate(const stream_list& pos_list_) const {
   assert( left_!= NULL && right_ != NULL );
   bool result_ = left_->evaluate(pos_list_) && right_->evaluate(pos_list_);
   if (!result_)
   {
      left_->reset();
      right_->reset();
   }
   return result_;
} // AndExpression::evaluate

MatchType AndExpression::match(const stream_list& pos_list_) const
{
    assert( left_!= NULL && right_ != NULL );
    MatchType left_match_ = left_->match(pos_list_);
    MatchType right_match_ = right_->match(pos_list_);

    std::copy(right_match_.begin(), right_match_.end(), std::back_inserter(left_match_));

    return left_match_;
}

void AndExpression::match(const stream_list& pos_list_, MatchType & left_match_) const
{
    assert( left_!= NULL && right_ != NULL );

    left_->match(pos_list_, left_match_);
    MatchType right_match_;
    right_->match(pos_list_, right_match_);

    std::copy(right_match_.begin(), right_match_.end(), std::back_inserter(left_match_));
}

std::size_t AndExpression::childrenCount() const {
   return 2;
} // AndExpression::childrenCount

Expression* AndExpression::child ( std::size_t index ) {
   if ( index > 1 ) return NULL;

   assert( left_!= NULL && right_ != NULL );
   return index == 0 ? left_ : right_;
} // AndExpression::child

void AndExpression::reset() {
   assert( left_!= NULL && right_ != NULL );
   left_->reset(); right_->reset();
} // AndExpression::reset

Expression::Kind::Type AndExpression::kind() const {
   return Expression::Kind::And;
} // AndExpression::kind

AndExpression::~AndExpression() {
   assert( left_!= NULL && right_ != NULL );
   delete left_; delete right_;
} // AndExpression::~AndExpression

// class OrExpression

const std::wstring OrExpression::name ( L"OR" );

OrExpression::OrExpression ( Expression* left, Expression* right )
   : left_( left ), right_( right )
{
   assert( left_!= NULL && right_ != NULL );
} // OrExpression::OrExpression

const std::wstring& OrExpression::sought() const {
   return OrExpression::name;
} // OrExpression::sought

bool OrExpression::evaluate(const stream_list& pos_list_) const {
    assert( left_!= NULL && right_ != NULL );
    bool leftResult = left_->evaluate(pos_list_);
    bool rightResult = rightResult = right_->evaluate(pos_list_);

    if (!leftResult)
        left_->reset();
    if (!rightResult)
        right_->reset();

    return leftResult || rightResult;
} // OrExpression::evaluate

MatchType OrExpression::match(const stream_list& pos_list_) const
{
    assert( left_!= NULL && right_ != NULL );
    MatchType left_match_ = left_->match(pos_list_);
    MatchType right_match_ = right_->match(pos_list_);

    std::copy(right_match_.begin(), right_match_.end(), std::back_inserter(left_match_));

    return left_match_;
}

void OrExpression::match(const stream_list& pos_list_, MatchType & left_match_) const
{
    assert( left_!= NULL && right_ != NULL );
    left_->match(pos_list_, left_match_);
    MatchType right_match_;
    right_->match(pos_list_, right_match_);

    std::copy(right_match_.begin(), right_match_.end(), std::back_inserter(left_match_));
}

std::size_t OrExpression::childrenCount() const {
   return 2;
} // OrExpression::childrenCount

Expression* OrExpression::child ( std::size_t index ) {
   if ( index > 1 ) return NULL;

   assert( left_!= NULL && right_ != NULL );
   return index == 0 ? left_ : right_;
} // OrExpression::child

void OrExpression::reset() {
   assert( left_!= NULL && right_ != NULL );
   left_->reset(); right_->reset();
} // OrExpression::reset

Expression::Kind::Type OrExpression::kind() const {
   return Expression::Kind::Or;
} // OrExpression::kind

OrExpression::~OrExpression() {
   assert( left_!= NULL && right_ != NULL );
   delete left_; delete right_;
} // OrExpression::~OrExpression
//------------------------------------------------------------------------------------------------------------------

const std::wstring NearExpression::name ( L"NEAR" );

inline bool in_range(long long pos_, long long start_, long long end_)
{
   return (start_ <= pos_) && (pos_ <= end_);
}

NearExpression::NearExpression( Expression* left, Expression* right, int distance)
   : left_( left ), right_( right ), distance_(distance)
{
   assert( left_!= NULL && right_ != NULL);
}

void NearExpression::check_distance(const stream_list& pos_list_, MatchType & result_) const
{
    assert( left_!= NULL && right_ != NULL );
    MatchType left_match_;
    left_->match(pos_list_, left_match_);
    MatchType right_match_;
    right_->match(pos_list_, right_match_);

    std::sort(left_match_.begin(), left_match_.end(), ComparePos);

    stream_list::const_iterator pos_it = pos_list_.begin();

    for (MatchType::const_iterator it_left = left_match_.begin(); it_left != left_match_.end(); ++it_left)
    {
        while (it_left->_pos > pos_it->_end_pos)
            ++pos_it;

        unsigned long long left_start = it_left->_pos - distance_;
        if (left_start < pos_it->_start_pos)
            left_start = pos_it->_start_pos;

        unsigned long long left_end = it_left->_pos - 1;
        if (left_end < pos_it->_start_pos)
            left_end = pos_it->_start_pos;

        unsigned long long right_start = it_left->_pos + it_left->_size;
        if (right_start > pos_it->_end_pos)
            right_start = pos_it->_end_pos;

        unsigned long long right_end = it_left->_pos + it_left->_size + distance_ - 1; 
        if (right_end > pos_it->_end_pos)
            right_end = pos_it->_end_pos;

        for (MatchType::const_iterator it_right = right_match_.begin(); it_right != right_match_.end(); ++it_right)
        {
            if (in_range(it_right->_pos + it_right->_size - 1, left_start, left_end))
            {
                result_.push_back(Matched(it_right->_pos, it_left->_pos - it_right->_pos + it_left->_size));
            }
            else if (in_range(it_right->_pos, right_start, right_end))
            {
                result_.push_back(Matched(it_left->_pos, it_right->_pos - it_left->_pos + it_right->_size));
            }
        }
    }
}

const std::wstring& NearExpression::sought() const
{
   return NearExpression::name;
} 

bool NearExpression::evaluate(const stream_list& pos_list_) const
{
    assert( left_!= NULL && right_ != NULL );
    MatchType result_;
    check_distance(pos_list_, result_);

    if (result_.empty())
        return false;
    else
        return true;
}

MatchType NearExpression::match(const stream_list& pos_list_) const
{
   MatchType result;
   check_distance(pos_list_, result);
   return result;
}

void NearExpression::match(const stream_list& pos_list_, MatchType & result) const
{
    check_distance(pos_list_, result);
}

std::size_t NearExpression::childrenCount() const
{
   return 2;
}

Expression* NearExpression::child ( std::size_t index )
{
   if ( index > 1 ) return NULL;

   assert( left_!= NULL && right_ != NULL );
   return index == 0 ? left_ : right_;
}

void NearExpression::reset()
{
   assert( left_!= NULL && right_ != NULL );
   left_->reset(); right_->reset();
}

Expression::Kind::Type NearExpression::kind() const
{
   return Expression::Kind::Near;
}

NearExpression::~NearExpression()
{
   assert( left_!= NULL && right_ != NULL );
   delete left_; delete right_;
} 

//------------------------------------------------------------------------------------------------------------------
// class SoughtExpression

SoughtExpression::SoughtExpression ( const std::wstring& sought ) : 
    sought_( sought )
    , preMatch_( false )
    , _u_pos()
{
} 

const std::wstring& SoughtExpression::sought() const 
{
   return sought_;
} 

bool SoughtExpression::evaluate(const stream_list& pos_list_) const 
{
   return preMatch_;
}

MatchType SoughtExpression::match(const stream_list& pos_list_) const
{
    return _u_pos;
}

void SoughtExpression::match(const stream_list& pos_list_, MatchType & result) const
{
    result = _u_pos;
}

void SoughtExpression::setMatch(unsigned long long local_pos_, unsigned long long global_pos_)
{
   if (local_pos_ - sought_.length() >= 0)
   {
       Matched matched_;
       matched_._pos = global_pos_ - sought_.length() +1;
       matched_._size = sought_.length();

       _u_pos.push_back(matched_);
       preMatch_ = true;
   }
} 

Expression::Kind::Type SoughtExpression::kind() const 
{
   return Expression::Kind::Sought;
} 

void SoughtExpression::rollBack() 
{
    _u_pos.pop_back();
    preMatch_ = !_u_pos.empty();
}

void SoughtExpression::reset() 
{
   preMatch_ = false;
   _u_pos.swap(MatchType());
} 

SoughtExpression::~SoughtExpression() 
{
} 
