//
// expression.hpp
//

#ifndef EXPRESSION_HPP_INCLUDED
#define EXPRESSION_HPP_INCLUDED

#include <cassert>
#include <cstddef>

#include <vector>
#include <list>
#include <string>

#include "result_params.h"
#include "matchstruct.hpp"

class Expression {
public:
    struct Kind { enum Type { Not, And, Or, Sought, Near }; };
public:
    virtual bool evaluate(const stream_list& pos_list_) const = 0;
    virtual MatchType match(const stream_list& pos_list_) const =0;
    virtual void match(const stream_list& pos_list_, MatchType & result) const =0;
    virtual void setMatch(unsigned long long local_pos_, unsigned long long global_pos_) {}
    virtual const std::wstring& sought() const = 0;
    virtual std::size_t childrenCount() const { return 0; }
    virtual Expression* child( std::size_t ) { return NULL; }
    virtual Kind::Type kind() const = 0;
    virtual void rollBack() {}
    virtual void reset() = 0;
    virtual void explain(std::wostream& ) =0;
    virtual void explain_with_tags(std::wostream& ) =0;
    virtual ~Expression() {}
}; // class Expression

class NotExpression : public Expression {
public:
    NotExpression( Expression* expression );
    bool evaluate(const stream_list& pos_list_) const;
    MatchType match(const stream_list& pos_list_) const;
    void match(const stream_list& pos_list_, MatchType & result) const;
    const std::wstring& sought() const;
    std::size_t childrenCount() const;
    Expression* child( std::size_t index );
    Kind::Type kind() const;
    void reset();
    virtual void explain(std::wostream& os_)
    {
        os_ << L"( NOT ";
        assert(child_);
        child_->explain(os_);	
        os_ << L" )";
    }
    virtual void explain_with_tags(std::wostream& os_)
    {
        os_ << L"<b>(</b> <font color=blue weight=700>NOT</font> ";

        assert(child_);
        child_->explain_with_tags(os_);	

        os_ << L" <b>)</b>";
    }

    ~NotExpression();
private:
    NotExpression( const NotExpression& );
    NotExpression& operator= ( const NotExpression& );
private:
    Expression* child_;
    static const std::wstring name;
}; // class NotExpression

class AndExpression : public Expression {
public:
    AndExpression( Expression* left, Expression* right );
    const std::wstring& sought() const;
    bool evaluate(const stream_list& pos_list_) const;
    MatchType match(const stream_list& pos_list_) const;
    void match(const stream_list& pos_list_, MatchType & result) const;
    std::size_t childrenCount() const;
    Expression* child( std::size_t index );
    void reset();
    virtual void explain(std::wostream& os_)
    {
        os_ << L"( ";

        assert(left_);
        left_->explain(os_);

        os_ << L" AND ";

        assert(right_);
        right_->explain(os_);	

        os_ << L" )";

    }
    virtual void explain_with_tags(std::wostream& os_)
    {
        os_ << L"<b>(</b> ";

        assert(left_);
        left_->explain_with_tags(os_);

        os_ << L" <font color=blue weight=700>AND</font> ";

        assert(right_);
        right_->explain_with_tags(os_);	

        os_ << L" <b>)</b>";

    }
    Kind::Type kind() const;
    ~AndExpression();
private:
    AndExpression( const AndExpression& );
    AndExpression& operator= ( const AndExpression& );
private:
    Expression* left_;
    Expression* right_;
    static const std::wstring name;
}; // class AndExpression

class OrExpression : public Expression {
public:
    OrExpression( Expression* left, Expression* right );
    const std::wstring& sought() const;
    bool evaluate(const stream_list& pos_list_) const;
    MatchType match(const stream_list& pos_list_) const;
    void match(const stream_list& pos_list_, MatchType & result) const;
    std::size_t childrenCount() const;
    Expression* child( std::size_t index );
    void reset();
    virtual void explain(std::wostream& os_)
    {
        os_ << L"( ";

        assert(left_);
        left_->explain(os_);

        os_ << L" OR ";

        assert(right_);
        right_->explain(os_);	

        os_ << L" )";

    }
    virtual void explain_with_tags(std::wostream& os_)
    {
        os_ << L"<b>(</b> ";

        assert(left_);
        left_->explain_with_tags(os_);

        os_ << L" <font color=blue weight=700>OR</font> ";

        assert(right_);
        right_->explain_with_tags(os_);	

        os_ << L" <b>)</b>";
    }


    Kind::Type kind() const;
    ~OrExpression();
private:
    OrExpression( const OrExpression& );
    OrExpression& operator= ( const OrExpression& );
private:
    Expression* left_;
    Expression* right_;
    static const std::wstring name;
}; // class OrExpression
//-----------------------------------------------------------------------------------------------------------------

class NearExpression : public Expression {
public:
    NearExpression( Expression* left, Expression* right, int distance);
    const std::wstring& sought() const;
    bool evaluate(const stream_list& pos_list_) const;
    MatchType match(const stream_list& pos_list_) const;
    void match(const stream_list& pos_list_, MatchType & result) const;
    void check_distance(const stream_list& pos_list_, MatchType & result) const;
    std::size_t childrenCount() const;
    Expression* child( std::size_t index );
    void reset();
    virtual void explain(std::wostream& os_)
    {
        os_ << L"( ";

        assert(left_);
        left_->explain(os_);

        os_ << L" NEAR/";
        os_ << distance_;
        os_ << L" ";

        assert(right_);
        right_->explain(os_);	

        os_ << L" )";

    }
    virtual void explain_with_tags(std::wostream& os_)
    {
        os_ << L"<b>(</b> ";

        assert(left_);
        left_->explain_with_tags(os_);

        os_ << L" <font color=blue weight=700>NEAR/";
        os_ << distance_;
        os_ << L" </font>";

        assert(right_);
        right_->explain_with_tags(os_);	

        os_ << L" <b>)</b>";
    }

    Kind::Type kind() const;
    ~NearExpression();
private:
    NearExpression( const NearExpression& );
    NearExpression& operator= ( const NearExpression& );
private:
    Expression* left_;
    Expression* right_;
    int distance_;
    static const std::wstring name;
}; // class NearExpression

//-----------------------------------------------------------------------------------------------------------------
class SoughtExpression : public Expression {
public:
    SoughtExpression( const std::wstring& sought );
    const std::wstring& sought() const;
    bool evaluate(const stream_list& pos_list_) const;
    MatchType match(const stream_list& pos_list_) const;
    void match(const stream_list& pos_list_, MatchType & result) const;
    void setMatch(unsigned long long local_pos_, unsigned long long global_pos_);
    Kind::Type kind() const;
    void rollBack();
    void reset();
    virtual void explain(std::wostream& os_)
    {
        //os_ << Convert(sought_);
    }
    virtual void explain_with_tags(std::wostream& os_)
    {
       /* std::vector<wchar_t> vec(sought_.getBuffer(), sought_.getBuffer() + sought_.length());
        for(std::vector<wchar_t>::const_iterator it_ = vec.begin(); it_ != vec.end(); ++it_)
        {
            if (*it_ == L'<')
                os_ << L"&lt;";
            else if (*it_ == L'>')
                os_ << L"&gt;";
            else if (*it_ == L'&')
                os_ << L"&amp;";
            else
                os_ << *it_;
        }*/
    }

    ~SoughtExpression();
private:
   /* inline std::wstring Convert(const ::std::wstring & str)
    {
        std::vector<wchar_t> vec(str.getBuffer(), str.getBuffer() + str.length());
        if (vec.empty())
            return std::wstring();
        else
            return std::wstring( & vec.at(0), vec.size());
    }*/

    SoughtExpression( const SoughtExpression& );
    SoughtExpression& operator= ( const SoughtExpression& );
private:
    bool preMatch_;
    const std::wstring sought_;

    MatchType _u_pos;
}; // class SoughtExpression

#endif // EXPRESSION_HPP_INCLUDED
