
#include <cctype>
#include <queue>

#include "parser.hpp"
#include "matcher.hpp"
#include "acmachine.hpp"

BooleanMatcher::BooleanMatcher( ExpressionParser& parser, bool wholeWord ) : 
parser_( parser )
, wholeWord_( wholeWord )
, current_lookbehind_position_(0)
, local_pos_(0)
, global_pos_(0)
, has_not_(false)
, pos_list_()
{
    const int maxSoughtLength = initializeMachine();
    lookbehind_buffer_.resize( maxSoughtLength + 1, L' ' );
} 

BooleanMatcher::~BooleanMatcher()
{
    for (stream_list::const_iterator it_ = pos_list_.begin(); it_ != pos_list_.end(); ++it_)
    {
        delete it_->_params;
    }
}
int BooleanMatcher::initializeMachine() 
{
    std::queue< Expression* > expressionsQueue;
    expressionsQueue.push( parser_.get() );

    int maxSoughtLength = 0;

    // Recursively extract all SougntExpession instances from search query

    while ( !expressionsQueue.empty() ) 
    {
        Expression* expression = expressionsQueue.front();
        expressionsQueue.pop();

        for ( std::size_t index = 0; index < expression->childrenCount(); ++index )
            expressionsQueue.push( expression->child( index ) );

        if (expression->kind() == Expression::Kind::Not)
            has_not_ = true;

        if ( expression->kind() == Expression::Kind::Sought )
        {
            machine_.addPattern( expression->sought(), expression );
            if ( expression->sought().length() > maxSoughtLength )
                maxSoughtLength = expression->sought().length();
        }
    }

    machine_.computeStates();

    return maxSoughtLength;
} 

void BooleanMatcher::addLookBehindCharacter( wchar_t character ) 
{
    if ( current_lookbehind_position_ == lookbehind_buffer_.size() )
        current_lookbehind_position_ = 0;

    lookbehind_buffer_[ current_lookbehind_position_ ] = character;
    ++current_lookbehind_position_;
} 

bool BooleanMatcher::lookBehindForSpace ( std::size_t position ) 
{
    if ( position < current_lookbehind_position_ )
    {
        return !iswalnum( lookbehind_buffer_[ current_lookbehind_position_ - position - 1 ] );
    } 
    else 
    {
        position = lookbehind_buffer_.size() - ( position - current_lookbehind_position_ ) - 1;
        return !iswalnum( lookbehind_buffer_[ position ] );
    }
} 

void BooleanMatcher::checkLastMatch( wchar_t character )
{
    // Check if last match wasn't a separate word

    const bool isSpace = !iswalnum( character );

    while ( !last_match_.empty() ) 
    {
        Expression* expression = last_match_.front();

        if ( !isSpace && local_pos_ > 0)
            expression->rollBack();

        last_match_.pop();
    }
} 

void BooleanMatcher::fill_pos_list(const stream_range& element_)
{
    pos_list_.push_back(element_);
}

const stream_list & BooleanMatcher::get_pos_list() const
{
    return pos_list_;
}

bool BooleanMatcher::feedCharacter ( wchar_t character ) 
{
    if ( wholeWord_ ) 
    {
        addLookBehindCharacter( character );
        checkLastMatch( character );
    }

    if ( !parser_.isCaseSensitive() )
        character = tolower( character );

    const bool hasMatch = machine_.feedCharacter( character);

    if ( hasMatch )
    {
        updateMatches();
    }

    ++local_pos_;
    ++global_pos_;

    return hasMatch;
} 

void BooleanMatcher::updateMatches() 
{
    const MatchList& matchList = machine_.getCurrentMatch();

    for (MatchList::const_iterator match = matchList.begin(); match != matchList.end(); ++match)
    {
        if ( !wholeWord_ || lookBehindForSpace( match->second->sought().length() ) ) {
            last_match_.push( match->second );
            match->second->setMatch(local_pos_, global_pos_);
        }
    }
} 

void BooleanMatcher::reset() 
{
    parser_.get()->reset();

    local_pos_ = 0;
    global_pos_ = 0;

    for (stream_list::const_iterator it_ = pos_list_.begin(); it_ != pos_list_.end(); ++it_)
    {
        delete it_->_params;
    }
    pos_list_.clear();

    while (!last_match_.empty())
        last_match_.pop();

    std::fill( lookbehind_buffer_.begin(), lookbehind_buffer_.end(), L' ' );
    current_lookbehind_position_ = 0;
} 

void BooleanMatcher::clear()
{
    local_pos_ = 0;

    std::fill( lookbehind_buffer_.begin(), lookbehind_buffer_.end(), L' ' );
    current_lookbehind_position_ = 0;
}
