
#include <iostream>
#include <cctype>

#include "parser.hpp"

// class KeyWords

KeyWords::KeyWords() {
   map_[ L"OR" ] = ExpressionParserImpl::OR;
   map_[ L"AND" ] = ExpressionParserImpl::AND;
   map_[ L"NOT" ] = ExpressionParserImpl::NOT;
   map_[ L"NEAR"] = ExpressionParserImpl::NEAR1;
} // KeyWords::KeyWords

int KeyWords::getToken ( const std::wstring& word ) const {
   KeyWordMap::const_iterator entry = map_.find( word );
   return entry != map_.end() ? entry->second : ExpressionParserImpl::Sought;
} // KeyWords::getToken

// class ExpressionParser

const KeyWords ExpressionParser::keywords_;

ExpressionParser::ExpressionParser ( const wchar_t* expression, std::size_t expressionSize, bool caseSensitive )
   : expression_str_(expression, expressionSize) 
   , charIt_(expression_str_.begin())
   , state_( State::Token )
   , quoted_( false )
{
   caseSensitive_ = caseSensitive;

   if ( yyparse() != 0 ) {
      throw std::runtime_error("expression parse error");
   }
} // ExpressionParser::ExpressionParser

ExpressionParser::~ExpressionParser() {
   delete expression_;
} // ExpressionParser::~ExpressionParser

void ExpressionParser::getSoughtsList ( SoughtsList& list ) const {
   list.clear();

   ExpressionsQueue queue;
   queue.push( expression_);

   while ( !queue.empty() ) {
      Expression* expression = queue.front(); queue.pop();

      for ( std::size_t index = 0; index < expression->childrenCount(); ++index )
         queue.push( expression->child( index ) );

      if ( expression->kind() == Expression::Kind::Sought )
         list.insert( expression->sought() );
   }
} // ExpressionParser::getSoughtsList

ExpressionParser::State::Type ExpressionParser::readQuoted() {
   ++charIt_; // Opening quote

   while ( charIt_ != expression_str_.end() && *charIt_ != QUOTE  ) {
      if ( *charIt_ == SLASH ) {
         if ( readEscaped() == State::Error )
            return State::Error;
      } else {
         word_ += *charIt_;
      }
      ++charIt_;
   }

   if ( charIt_ == expression_str_.end())
      return State::Error;

   ++charIt_; // Ending quote
   quoted_ = true;

   return state_;
} // ExpressionParser::readQuoted

ExpressionParser::State::Type ExpressionParser::readEscaped() {
  
   ++charIt_; //Skip slash
   if ( charIt_ == expression_str_.end() )
      return State::Error;

   if ( *charIt_ != QUOTE && *charIt_ != SLASH )
      return State::Error;

   word_ += *(charIt_++);

   return state_;
} // ExpressionParser::readEscaped

ExpressionParser::State::Type ExpressionParser::readCharacter() {
    if (charIt_ == expression_str_.end())
        return ExpressionParser::State::Error;

   wchar_t c = *(charIt_++);
   if ( iswspace( c ) ) {
      if (!word_.empty() && state_ != State::Near && state_ != State::Back_Slash)
      {
         return State::Ready;
      }
      if (!word_.empty() && state_ == State::Back_Slash)
         return State::Distance;
      return state_;
   }

   word_ += c;
   return state_;
} // ExpressionParser::readCharacter

int ExpressionParser::parseWord ( yy_ExpressionParserImpl_stype* element ) {
   int token = ExpressionParserImpl::Sought;

   if ( !quoted_ )
      token = keywords_.getToken( word_ );

   if ( token == ExpressionParserImpl::Sought )
      element->sought_ = new std::wstring ( word_ );

   if ( token == ExpressionParserImpl::NEAR1)
      state_ = State::Near;
   else 
      state_ = State::Token;

   quoted_ = false;
   word_.clear();

   return token;
} // ExpressionParser::parseWord


int ExpressionParser::yylex ( yy_ExpressionParserImpl_stype* element ) {
   element->expression_ = NULL;
   element->sought_ = NULL;

   while ( charIt_ != expression_str_.end() ) {
      switch ( *charIt_ ) {
         case QUOTE: state_ = readQuoted(); break;
         case OPEN: 
             ++charIt_;
             return ExpressionParserImpl::OPEN;

         case CLOSE:
             ++charIt_;
             return ExpressionParserImpl::CLOSE;
         default: state_ = readCharacter();
      }
      if ( state_ == State::Near )
      {
         if (word_[0] == BACK_SLASH)
         {
            state_ = State::Back_Slash;
            word_.clear();

            return ExpressionParserImpl::BACK_SLASH;
         }
      }
      if ( state_ == State::Distance)
      {
         if (word_.empty())
            return ExpressionParserImpl::YY_ExpressionParserImpl_NULL_TOKEN;
         else
         {
            element->size_type_ = _wtoi(word_.c_str());
            word_.clear();
            state_ = State::Token;
            return ExpressionParserImpl::TOK_SIZE;
         }
      }
      
      if ( state_ == State::Ready )
         return parseWord( element );

      if ( state_ == State::Error )
         return ExpressionParserImpl::YY_ExpressionParserImpl_NULL_TOKEN;
   }

   if ( !word_.empty() )
      return parseWord( element );

   return 0;
} // ExpressionParser::yylex

void ExpressionParser::yyerror ( char* /*message*/ ) {
} // ExpressionParser::yyerror
